#ifndef _IPXE_PRIVKEY_H
#define _IPXE_PRIVKEY_H

/** @file
 *
 * Private key
 *
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <ipxe/asn1.h>

extern struct asn1_cursor private_key;

#endif /* _IPXE_PRIVKEY_H */
