#ifndef _IPXE_ISQRT_H
#define _IPXE_ISQRT_H

/** @file
 *
 * Integer square root
 *
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

extern unsigned long isqrt ( unsigned long value );

#endif /* _IPXE_ISQRT_H */
