/** @file
 *
 * "uk" keyboard mapping
 *
 * This file is automatically generated; do not edit
 *
 */

FILE_LICENCE ( PUBLIC_DOMAIN );

#include <ipxe/keymap.h>

/** "uk" keyboard mapping */
struct key_mapping uk_mapping[] __keymap = {
	{ 0x22, 0x40 },	/* '"' => '@' */
	{ 0x40, 0x22 },	/* '@' => '"' */
	{ 0x5c, 0x23 },	/* '\\' => '#' */
	{ 0x7c, 0x7e },	/* '|' => '~' */
};
