#ifndef _IPXE_DER_H
#define _IPXE_DER_H

/** @file
 *
 * DER image format
 *
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <ipxe/image.h>

extern struct image_type der_image_type __image_type ( PROBE_NORMAL );

#endif /* _IPXE_DER_H */
