#ifndef _IPXE_ROOTCERT_H
#define _IPXE_ROOTCERT_H

/** @file
 *
 * Root certificate store
 *
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <ipxe/x509.h>

extern struct x509_root root_certificates;

#endif /* _IPXE_ROOTCERT_H */
