#ifndef _BITS_ERRFILE_H
#define _BITS_ERRFILE_H

/** @file
 *
 * ARM-specific error file identifiers
 *
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

/**
 * @addtogroup errfile Error file identifiers
 * @{
 */

/** @} */

#endif /* _BITS_ERRFILE_H */
