#ifndef _USR_SYNC_H
#define _USR_SYNC_H

/** @file
 *
 * Wait for pending operations to complete
 *
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

extern int sync ( unsigned long timeout );

#endif /* _USR_SYNC_H */
