#ifndef _BITS_UMALLOC_H
#define _BITS_UMALLOC_H

/** @file
 *
 * x86-specific user memory allocation API implementations
 *
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <ipxe/memtop_umalloc.h>

#endif /* _BITS_UMALLOC_H */
