#ifndef CONFIG_TIME_H
#define CONFIG_TIME_H

/** @file
 *
 * Time API configuration
 *
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <config/defaults.h>

#include <config/local/time.h>

#endif /* CONFIG_TIME_H */
