#ifndef _USR_NSLOOKUP_H
#define _USR_NSLOOKUP_H

/** @file
 *
 * Standalone name resolution
 *
 */

FILE_LICENCE ( GPL2_OR_LATER );

extern int nslookup ( const char *name, const char *setting_name );

#endif /* _USR_NSLOOKUP_H */
